from __future__ import annotations

import dataclasses
from typing import Any, Literal

from . import _pretty


class TaskError(Exception):
    """Exception raised when an action or exec command fails."""

    task: Task
    """Associated task."""

    def __init__(self, task: Task):
        self.task = task

    def __str__(self) -> str:
        return f'task error: {self.task}'


@dataclasses.dataclass(frozen=True)
class Task:
    """A task holds the results of Juju running an action or exec command on a single unit."""

    id: str
    """Task ID of the action, for use with ``juju show-task``."""

    status: Literal['aborted', 'cancelled', 'completed', 'error', 'failed']
    """Status of the action (Juju operation). Typically "completed" or "failed"."""

    results: dict[str, Any] = dataclasses.field(default_factory=dict)
    """Results of the action provided by the charm.

    This excludes the special "return-code", "stdout", and "stderr" keys
    inserted by Juju; those values are provided by separate attributes.
    """

    return_code: int = 0
    """Return code from executing the charm action hook."""

    stdout: str = ''
    """Stdout printed by the action hook."""

    stderr: str = ''
    """Stderr printed by the action hook."""

    message: str = ''
    """Failure message, if the charm provided a message when it failed the action."""

    log: list[str] = dataclasses.field(default_factory=list)
    """List of messages logged by the action hook."""

    def __str__(self) -> str:
        details: list[str] = []
        if self.results:
            details.append(f'Results: {self.results}')
        if self.stdout:
            details.append(f'Stdout:\n{self.stdout}')
        if self.stderr:
            details.append(f'Stderr:\n{self.stderr}')
        if self.message:
            details.append(f'Message: {self.message}')
        if self.log:
            log_str = '\n'.join(self.log)
            details.append(f'Log:\n{log_str}')
        s = f'Task {self.id}: status {self.status!r}, return code {self.return_code}'
        if details:
            s += ', details:\n' + '\n'.join(details)
        return s

    def __repr__(self) -> str:
        return _pretty.dump(self)

    @classmethod
    def _from_dict(cls, d: dict[str, Any]) -> Task:
        results: dict[str, Any] = d.get('results') or {}
        return_code = results.pop('return-code', 0)
        stdout = results.pop('stdout', '')
        stderr = results.pop('stderr', '')
        return cls(
            id=d['id'],
            status=d['status'],
            results=results,
            return_code=return_code,
            stdout=stdout,
            stderr=stderr,
            message=d.get('message') or '',
            log=d.get('log') or [],
        )

    @property
    def success(self) -> bool:
        """Whether the action was successful."""
        return self.status == 'completed' and self.return_code == 0

    def raise_on_failure(self):
        """If task was not successful, raise a :class:`TaskError`."""
        if not self.success:
            raise TaskError(self)



from __future__ import annotations

from collections.abc import Iterable

from .statustypes import Status


def all_active(status: Status, *apps: str) -> bool:
    """Report whether all apps and units in *status* (or in *apps* if provided) are "active".

    Examples::

        # Use the callable directly to wait for all apps in status to be active.
        juju.wait(jubilant.all_active)

        # Use a lambda to wait for all apps specified (blog, mysql) to be active.
        juju.wait(lambda status: jubilant.all_active(status, 'blog', 'mysql'))

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested. If an app is not
            present in ``status.apps``, returns False.
    """
    return _all_statuses_are('active', status, apps)


def all_blocked(status: Status, *apps: str) -> bool:
    """Report whether all apps and units in *status* (or in *apps* if provided) are "blocked".

    See :func:`all_active` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested. If an app is not
            present in ``status.apps``, returns False.
    """
    return _all_statuses_are('blocked', status, apps)


def all_error(status: Status, *apps: str) -> bool:
    """Report whether all apps and units in *status* (or in *apps* if provided) are "error".

    See :func:`all_active` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested. If an app is not
            present in ``status.apps``, returns False.
    """
    return _all_statuses_are('error', status, apps)


def all_maintenance(status: Status, *apps: str) -> bool:
    """Report whether all apps and units in *status* (or in *apps* if provided) are "maintenance".

    See :func:`all_active` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested. If an app is not
            present in ``status.apps``, returns False.
    """
    return _all_statuses_are('maintenance', status, apps)


def all_waiting(status: Status, *apps: str) -> bool:
    """Report whether all apps and units in *status* (or in *apps* if provided) are "waiting".

    See :func:`all_active` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested. If an app is not
            present in ``status.apps``, returns False.
    """
    return _all_statuses_are('waiting', status, apps)


def any_active(status: Status, *apps: str) -> bool:
    """Report whether any app or unit in *status* (or in *apps* if provided) is "active".

    See :func:`any_error` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested.
    """
    return _any_status_is('active', status, apps)


def any_blocked(status: Status, *apps: str) -> bool:
    """Report whether any app or unit in *status* (or in *apps* if provided) is "blocked".

    See :func:`any_error` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested.
    """
    return _any_status_is('blocked', status, apps)


def any_error(status: Status, *apps: str) -> bool:
    """Report whether any app or unit in *status* (or in *apps* if provided) is "error".

    Examples::

        # Use the callable directly to raise an error if any apps go into error.
        juju.wait(jubilant.all_active, error=jubilant.any_error)

        # Use a lambda to wait for any of the apps specified (blog, mysql) to go into error.
        juju.wait(
            jubilant.all_active,
            error=lambda status: jubilant.any_error(status, 'blog', 'mysql')),
        )

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested.
    """
    return _any_status_is('error', status, apps)


def any_maintenance(status: Status, *apps: str) -> bool:
    """Report whether any app or unit in *status* (or in *apps* if provided) is "maintenance".

    See :func:`any_error` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested.
    """
    return _any_status_is('maintenance', status, apps)


def any_waiting(status: Status, *apps: str) -> bool:
    """Report whether any app or unit in *status* (or in *apps* if provided) is "waiting".

    See :func:`any_error` for examples.

    Args:
        status: The status object being tested.
        apps: If provided, only these applications (and their units) are tested.
    """
    return _any_status_is('waiting', status, apps)


def all_agents_idle(status: Status, *apps: str) -> bool:
    """Report whether all unit agents in *status* (filtered to *apps* if provided) are "idle".

    Unlike the other ``all_*`` and ``any_*`` helpers, this method looks at the status of each
    Juju unit agent, not the workload's application or unit status.

    Examples::

        # Use the callable directly to wait for unit agents from all apps to be idle.
        juju.wait(jubilant.all_agents_idle)

        # Use a lambda to wait for unit agents only from specified apps (blog, mysql).
        juju.wait(lambda status: jubilant.all_agents_idle(status, 'blog', 'mysql'))

    Args:
        status: The status object being tested.
        apps: If provided, only the unit agents of units from these applications are tested.
            If an app is not present in ``status.apps``, returns False.
    """
    return _all_agent_statuses_are('idle', status, apps)


def _all_statuses_are(expected: str, status: Status, apps: Iterable[str]) -> bool:
    if not apps:
        apps = status.apps

    for app in apps:
        app_info = status.apps.get(app)
        if app_info is None:
            return False
        if app_info.app_status.current != expected:
            return False
        for unit_info in status.get_units(app).values():
            if unit_info.workload_status.current != expected:
                return False
    return True


def _any_status_is(expected: str, status: Status, apps: Iterable[str]) -> bool:
    if not apps:
        apps = status.apps

    for app in apps:
        app_info = status.apps.get(app)
        if app_info is None:
            continue
        if app_info.app_status.current == expected:
            return True
        for unit_info in status.get_units(app).values():
            if unit_info.workload_status.current == expected:
                return True
    return False


def _all_agent_statuses_are(expected: str, status: Status, apps: Iterable[str]) -> bool:
    if not apps:
        apps = status.apps

    for app in apps:
        app_info = status.apps.get(app)
        if app_info is None:
            return False
        for unit_info in status.get_units(app).values():
            if unit_info.juju_status.current != expected:
                return False
    return True



from __future__ import annotations

import contextlib
import secrets
from typing import Generator

from ._juju import Juju


@contextlib.contextmanager
def temp_model(keep: bool = False, controller: str | None = None) -> Generator[Juju]:
    """Context manager to create a temporary model for running tests in.

    This creates a new model with a random name in the format ``jubilant-abcd1234``, and destroys
    it and its storage when the context manager exits.

    Provides a :class:`Juju` instance to operate on.

    Args:
        keep: If true, keep the created model around when the context manager exits.
        controller: Name of controller where the temporary model will be added.
    """
    juju = Juju()
    model = 'jubilant-' + secrets.token_hex(4)  # 4 bytes (8 hex digits) should be plenty
    juju.add_model(model, controller=controller)
    try:
        yield juju
    finally:
        if not keep:
            juju.destroy_model(model, destroy_storage=True, force=True)



from __future__ import annotations

import dataclasses
import difflib
from collections.abc import Generator, Sequence
from typing import cast

_MAX_VALUE = 150
_INDENT = 2
_SIMPLE_TYPES = (type(None), bool, int, float, str)


def dump(value: object, indent: str = '') -> str:
    """Pretty-print a value with special cases for dataclasses, lists, and dicts.

    If *value*'s fields or sub-values are simple scalar types and the result fits in a limited
    number of characters, use a single-line format. Otherwise, split onto multiple lines and
    indent each line.

    For dataclasses, omit fields that are set to their default (or default_factory return value).
    """
    sub_indent = indent + ' ' * _INDENT

    if dataclasses.is_dataclass(value):
        fields: list[str] = []
        is_simple = True
        for field in dataclasses.fields(value):
            v = getattr(value, field.name)
            if field.default is not dataclasses.MISSING and v == field.default:
                continue
            if field.default_factory is not dataclasses.MISSING and v == field.default_factory():
                continue
            v_str = dump(v, sub_indent)
            fields.append(f'{field.name}={v_str}')
            is_simple = is_simple and isinstance(v, _SIMPLE_TYPES)

        class_name = value.__class__.__name__  # type: ignore
        if is_simple:
            # If one of the dumped fields above has been split onto multiple lines due to it
            # exceeding _MAX_VALUE, this will definitely not fit in _MAX_VALUE.
            single_line = f'{class_name}({", ".join(fields)})'
            if len(single_line) <= _MAX_VALUE:
                return single_line

        lines_str = '\n'.join(sub_indent + f + ',' for f in fields)
        return f'{class_name}(\n{lines_str}\n{indent})'

    elif isinstance(value, list):
        value = cast('list[object]', value)
        is_simple = all(isinstance(v, _SIMPLE_TYPES) for v in value)
        if is_simple:
            single_line = repr(value)
            if len(single_line) <= _MAX_VALUE:
                return single_line

        items = [dump(v, sub_indent) for v in value]
        lines_str = '\n'.join(sub_indent + item + ',' for item in items)
        return f'[\n{lines_str}\n{indent}]'

    elif isinstance(value, dict):
        value = cast('dict[str, object]', value)
        is_simple = all(isinstance(v, _SIMPLE_TYPES) for v in value.values())
        if is_simple:
            single_line = repr(value)
            if len(single_line) <= _MAX_VALUE:
                return single_line

        items = [f'{k!r}: {dump(v, sub_indent)}' for k, v in sorted(value.items())]
        lines_str = '\n'.join(sub_indent + item + ',' for item in items)
        return f'{{\n{lines_str}\n{indent}}}'

    else:
        return repr(value)


def gron(value: object, prefix: str = '') -> Generator[str]:
    """Yield gron-style lines of all fields within value, recursively.

    This handles dataclasses, lists, and dicts. It's inspired by gron:
    https://github.com/tomnomnom/gron

    Example output for a :class:`Status` instance with ``prefix='status'`` (some lines omitted
    for brevity)::

        status.model.name = 'tt'
        status.model.controller = 'microk8s-localhost'
        status.model.version = '3.6.1'
        status.apps['database'].charm = 'local:database-0'
        status.apps['database'].charm_rev = 0
        status.apps['database'].base.name = 'ubuntu'
        status.apps['database'].base.channel = '22.04'
        status.apps['database'].app_status.current = 'active'
        status.apps['database'].app_status.message = 'relation-created: added new secret'
        status.apps['database'].app_status.since = '24 Feb 2025 16:59:43+13:00'
        status.apps['database'].relations['db'][0].related_app = 'webapp'
        status.apps['database'].relations['db'][0].interface = 'dbi'
        status.apps['database'].relations['db'][0].scope = 'global'
        ...
    """
    if dataclasses.is_dataclass(value):
        for field in dataclasses.fields(value):
            v = getattr(value, field.name)
            if field.default is not dataclasses.MISSING and v == field.default:
                continue
            if field.default_factory is not dataclasses.MISSING and v == field.default_factory():
                continue
            yield from gron(v, f'{prefix}.{field.name}')

    elif isinstance(value, list):
        value = cast('list[object]', value)
        for i, v in enumerate(value):
            yield from gron(v, f'{prefix}[{i}]')

    elif isinstance(value, dict):
        value = cast('dict[str, object]', value)
        for k, v in sorted(value.items()):
            yield from gron(v, f'{prefix}[{k!r}]')

    else:
        yield f'{prefix} = {value!r}'


def diff(seq1: Sequence[str], seq2: Sequence[str]) -> Generator[str]:
    """Compare seq1 and seq1; yield lines that have been removed, changed, or added.

    Example::

        - .apps['database'].app_status.current = 'active'
        + .apps['database'].app_status.current = 'waiting'
        - .apps['database'].relations['db'][0].scope = 'global'
        - .apps['database'].relations['db'][1].related_app = 'dummy'
        - .apps['database'].relations['db'][1].interface = 'xyz'
        - .apps['database'].relations['db'][1].scope = 'foobar'
        + .apps['database'].relations['db'][0].scope = 'testy'
    """
    matcher = difflib.SequenceMatcher(None, seq1, seq2)
    for tag, i1, i2, j1, j2 in matcher.get_opcodes():
        if tag in {'replace', 'delete'}:
            for line in seq1[i1:i2]:
                yield '- ' + line
        if tag in {'replace', 'insert'}:
            for line in seq2[j1:j2]:
                yield '+ ' + line



"""Jubilant is a Pythonic wrapper around the Juju CLI for writing charm integration tests."""

from . import statustypes
from ._all_any import (
    all_active,
    all_agents_idle,
    all_blocked,
    all_error,
    all_maintenance,
    all_waiting,
    any_active,
    any_blocked,
    any_error,
    any_maintenance,
    any_waiting,
)
from ._juju import CLIError, ConfigValue, Juju, SecretURI, WaitError
from ._task import Task, TaskError
from ._test_helpers import temp_model
from .statustypes import Status

__all__ = [
    'CLIError',
    'ConfigValue',
    'Juju',
    'SecretURI',
    'Status',
    'Task',
    'TaskError',
    'WaitError',
    'all_active',
    'all_agents_idle',
    'all_blocked',
    'all_error',
    'all_maintenance',
    'all_waiting',
    'any_active',
    'any_blocked',
    'any_error',
    'any_maintenance',
    'any_waiting',
    'statustypes',
    'temp_model',
]

__version__ = '1.2.0'
